(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[9650],{44085:function(e,i,n){"use strict";n.r(i);var r=n(23256),s=n(44359),t=n(20440),o=n(41664),c=n.n(o),u=n(11163),d=n(18459),a=n(84247),l=n(67971),h=n(57418),v=n(85893);i.default=function(){var e,i=(0,u.useRouter)().query.id,n=void 0===i?"":i,o=(0,l.QA)({id:Array.isArray(n)?n[0]:n,verbose:!0},{skip:""===n}),x=o.data,f=o.isLoading;return e=o.isUninitialized||f?(0,v.jsx)(r.M5,{children:(0,v.jsx)(s.$,{})}):x&&0!==(null===x||void 0===x?void 0:x.items.length)?(0,v.jsx)(h.Z,{data:null===x||void 0===x?void 0:x.items[0]}):(0,v.jsx)(r.xv,{children:"404 no privacy request found"}),(0,v.jsxs)(a.Z,{title:"Privacy Requests - ".concat(n),children:[(0,v.jsxs)(r.X6,{fontSize:"2xl",fontWeight:"semibold",children:["Privacy Request",(0,v.jsx)(r.xu,{mt:2,mb:9,children:(0,v.jsxs)(t.aG,{fontWeight:"medium",fontSize:"sm",children:[(0,v.jsx)(t.gN,{children:(0,v.jsx)(t.At,{as:c(),href:d.Jw,children:"Privacy Requests"})}),(0,v.jsx)(t.gN,{children:(0,v.jsx)(t.At,{isCurrentPage:!0,color:"complimentary.500",_hover:{textDecoration:"none"},children:"View Details"})})]})})]}),e]})}},44216:function(e,i,n){(window.__NEXT_P=window.__NEXT_P||[]).push(["/privacy-requests/[id]",function(){return n(44085)}])}},function(e){e.O(0,[5372,7601,2890,2350,440,8522,5657,6508,7418,9774,2888,179],(function(){return i=44216,e(e.s=i);var i}));var i=e.O();_N_E=i}]);