.ConnectionGrid_grid-row__pjx_I{border-top-width:.5px;border-color:var(--chakra-colors-blackAlpha-300);box-sizing:border-box}.ConnectionGrid_grid-item__HdQkZ{border-right-width:.5px;border-color:var(--chakra-colors-blackAlpha-300);box-sizing:border-box}.ConnectionGrid_grid-item__HdQkZ:nth-child(3n){border-right-width:0}