"""add security policy link and erasure time limits

Revision ID: 1739aa4a4ab7
Revises: 4c3693c289d0
Create Date: 2022-01-28 20:41:20.065432

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "1739aa4a4ab7"
down_revision = "4c3693c289d0"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "organizations", sa.Column("security_policy", sa.String(), nullable=True)
    )
    op.add_column("datasets", sa.Column("retention", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("organizations", "security_policy")
    op.drop_column("datasets", "retention")
    # ### end Alembic commands ###
