# Prefix
V1_URL_PREFIX = "/api/v1"

# Login URLs
LOGIN = "/login"

# Oauth Client URLs
TOKEN = "/oauth/token"

# User URLs
USERS = "/user"
USER_DETAIL = "/user/{user_id}"
USER_PASSWORD_RESET = "/user/{user_id}/reset-password"

# User Permission URLs
USER_PERMISSIONS = "/user/{user_id}/permission"
