"use strict";(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[2350],{82350:function(e,n,t){t.d(n,{r:function(){return f}});var r=t(12036),a=t(37178);t(38554),t(41706);var i=!1;!function(e){var n=new WeakMap}((function(e,n,t,r){var a="string"===typeof n?n.split("."):[n];for(r=0;r<a.length&&e;r+=1)e=e[a[r]];return void 0===e?t:e}));"undefined"===typeof window||!window.document||window.document.createElement;var o=function(e){return e?"":void 0},c=function(){for(var e=arguments.length,n=new Array(e),t=0;t<e;t++)n[t]=arguments[t];return n.filter(Boolean).join(" ")};["input:not([disabled])","select:not([disabled])","textarea:not([disabled])","embed","iframe","object","a[href]","area[href]","button:not([disabled])","[tabindex]","audio[controls]","video[controls]","*[tabindex]:not([aria-disabled])","*[contenteditable]"].join();function s(e){var n;return function(){if(e){for(var t=arguments.length,r=new Array(t),a=0;a<t;a++)r[a]=arguments[a];n=e.apply(this,r),e=null}return n}}s((function(e){return function(){e.condition,e.message}})),s((function(e){return function(){e.condition,e.message}}));Number.MIN_SAFE_INTEGER,Number.MAX_SAFE_INTEGER;Object.freeze(["base","sm","md","lg","xl","2xl"]);var l=t(67294);function u(){return u=Object.assign||function(e){for(var n=1;n<arguments.length;n++){var t=arguments[n];for(var r in t)Object.prototype.hasOwnProperty.call(t,r)&&(e[r]=t[r])}return e},u.apply(this,arguments)}var d=["spacing","children"],f=(0,a.Gp)((function(e,n){var t=(0,a.jC)("Switch",e),i=(0,a.Lr)(e),s=i.spacing,f=void 0===s?"0.5rem":s,h=i.children,m=function(e,n){if(null==e)return{};var t,r,a={},i=Object.keys(e);for(r=0;r<i.length;r++)t=i[r],n.indexOf(t)>=0||(a[t]=e[t]);return a}(i,d),p=(0,r.O)(m),b=p.state,_=p.getInputProps,v=p.getCheckboxProps,g=p.getRootProps,k=p.getLabelProps,w=l.useMemo((function(){return u({display:"inline-block",position:"relative",verticalAlign:"middle",lineHeight:0},t.container)}),[t.container]),E=l.useMemo((function(){return u({display:"inline-flex",flexShrink:0,justifyContent:"flex-start",boxSizing:"content-box",cursor:"pointer"},t.track)}),[t.track]),N=l.useMemo((function(){return u({userSelect:"none",marginStart:f},t.label)}),[f,t.label]);return l.createElement(a.m$.label,u({},g(),{className:c("chakra-switch",e.className),__css:w}),l.createElement("input",u({className:"chakra-switch__input"},_({},n))),l.createElement(a.m$.span,u({},v(),{className:"chakra-switch__track",__css:E}),l.createElement(a.m$.span,{__css:t.thumb,className:"chakra-switch__thumb","data-checked":o(b.isChecked),"data-hover":o(b.isHovered)})),h&&l.createElement(a.m$.span,u({className:"chakra-switch__label"},k(),{__css:N}),h))}))}}]);