.BorderGrid_grid-row__gzSVf{border-top-width:.5px;border-color:var(--chakra-colors-blackAlpha-300);box-sizing:border-box}.BorderGrid_grid-item__7gUTW{border-right-width:.5px;border-color:var(--chakra-colors-blackAlpha-300);box-sizing:border-box}.BorderGrid_grid-item__7gUTW:nth-child(3n){border-right-width:0}