"use strict";(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[9549],{29549:function(e,t,i){i.r(t),i.d(t,{NavLink:function(){return l},default:function(){return u}});var r=i(41664),n=i.n(r),c=i(11163),s=(i(67294),i(37178)),a=i(43894),o=i(85893),h=(0,s.Gp)((function(e,t){var i=e.disabled,r=e.href,n=e.isActive,c=e.rightIcon,s=e.title,h=e.onClick;return(0,o.jsx)(a.zx,{ref:t,as:"a",href:i?void 0:r,onClick:h,variant:"ghost",disabled:i,mr:4,colorScheme:n?"complimentary":"ghost",rightIcon:c,"data-testid":"nav-link-".concat(s),isActive:n,_active:{bg:"transparent"},children:s})})),f=i(28026),l=function(e){var t=e.title,i=e.href,r=e.disabled,s=e.rightIcon,a=e.exact,l=(0,c.useRouter)(),u=(0,f.qj)({href:i,router:l,exact:a});return r||!u.isCurrentZone?(0,o.jsx)(h,{disabled:r,href:u.href,isActive:u.isActive,rightIcon:s,title:t}):(0,o.jsx)(n(),{href:u.href,passHref:!0,children:(0,o.jsx)(h,{isActive:u.isActive,rightIcon:s,title:t})})},u=l}}]);