from fides.api.ops.service.connectors.sql_connector import PostgreSQLConnector


class TimescaleConnector(PostgreSQLConnector):
    """TimescaleDB Connector - This currently supports identical functionality to the PostgreSQL connector"""
