"""add fidesctl_meta to organization object

Revision ID: 150f234ef1de
Revises: edcd28ede1f7
Create Date: 2022-03-07 20:14:00.530408

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "150f234ef1de"
down_revision = "edcd28ede1f7"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("organizations", sa.Column("fidesctl_meta", sa.JSON(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("organizations", "fidesctl_meta")
    # ### end Alembic commands ###
