"use strict";(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[4],{30004:function(n,e,r){r.d(e,{Ee:function(){return g}});var t=r(37178);r(38554),r(41706);var a=!1;!function(n){var e=new WeakMap}((function(n,e,r,t){var a="string"===typeof e?e.split("."):[e];for(t=0;t<a.length&&n;t+=1)n=n[a[t]];return void 0===n?r:n}));"undefined"===typeof window||!window.document||window.document.createElement;["input:not([disabled])","select:not([disabled])","textarea:not([disabled])","embed","iframe","object","a[href]","area[href]","button:not([disabled])","[tabindex]","audio[controls]","video[controls]","*[tabindex]:not([aria-disabled])","*[contenteditable]"].join();function i(n){var e;return function(){if(n){for(var r=arguments.length,t=new Array(r),a=0;a<r;a++)t[a]=arguments[a];e=n.apply(this,t),n=null}return e}}i((function(n){return function(){n.condition,n.message}})),i((function(n){return function(){n.condition,n.message}}));Number.MIN_SAFE_INTEGER,Number.MAX_SAFE_INTEGER;Object.freeze(["base","sm","md","lg","xl","2xl"]);var o=r(67294),c=r(44697);function l(){return l=Object.assign||function(n){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var t in r)Object.prototype.hasOwnProperty.call(r,t)&&(n[t]=r[t])}return n},l.apply(this,arguments)}function u(n,e){if(null==n)return{};var r,t,a={},i=Object.keys(n);for(t=0;t<i.length;t++)r=i[t],e.indexOf(r)>=0||(a[r]=n[r]);return a}var s=["htmlWidth","htmlHeight","alt"],f=["fallbackSrc","fallback","src","srcSet","align","fit","loading","ignoreFallback","crossOrigin"],d=o.forwardRef((function(n,e){var r=n.htmlWidth,t=n.htmlHeight,a=n.alt,i=u(n,s);return o.createElement("img",l({width:r,height:t,ref:e,alt:a},i))})),g=(0,t.Gp)((function(n,e){var r=n.fallbackSrc,a=n.fallback,i=n.src,s=n.srcSet,g=n.align,b=n.fit,m=n.loading,h=n.ignoreFallback,v=n.crossOrigin,p=u(n,f),k=null!=m||h||void 0===r&&void 0===a,E=function(n){var e=n.loading,r=n.src,t=n.srcSet,a=n.onLoad,i=n.onError,l=n.crossOrigin,u=n.sizes,s=n.ignoreFallback,f=(0,o.useState)("pending"),d=f[0],g=f[1];(0,o.useEffect)((function(){g(r?"loading":"pending")}),[r]);var b=(0,o.useRef)(),m=(0,o.useCallback)((function(){if(r){h();var n=new Image;n.src=r,l&&(n.crossOrigin=l),t&&(n.srcset=t),u&&(n.sizes=u),e&&(n.loading=e),n.onload=function(n){h(),g("loaded"),null==a||a(n)},n.onerror=function(n){h(),g("failed"),null==i||i(n)},b.current=n}}),[r,l,t,u,a,i,e]),h=function(){b.current&&(b.current.onload=null,b.current.onerror=null,b.current=null)};return(0,c.a)((function(){if(!s)return"loading"===d&&m(),function(){h()}}),[d,m,s]),s?"loaded":d}(l({},n,{ignoreFallback:k})),w=l({ref:e,objectFit:b,objectPosition:g},k?p:function(n,e){var r={};return Object.keys(n).forEach((function(t){e.includes(t)||(r[t]=n[t])})),r}(p,["onError","onLoad"]));return"loaded"!==E?a||o.createElement(t.m$.img,l({as:d,className:"chakra-image__placeholder",src:r},w)):o.createElement(t.m$.img,l({as:d,src:i,srcSet:s,crossOrigin:v,loading:m,className:"chakra-image"},w))}))}}]);