CONSENT_REQUEST_VERIFICATION_TEMPLATE = "consent_request_verification.html"
SUBJECT_IDENTITY_VERIFICATION_TEMPLATE = "subject_identity_verification.html"
EMAIL_ERASURE_REQUEST_FULFILLMENT = "message_request_email_fulfillment.html"
PRIVACY_REQUEST_RECEIPT_TEMPLATE = "privacy_request_receipt.html"
PRIVACY_REQUEST_COMPLETE_DELETION_TEMPLATE = "privacy_request_complete_deletion.html"
PRIVACY_REQUEST_COMPLETE_ACCESS_TEMPLATE = "privacy_request_complete_access.html"
PRIVACY_REQUEST_ERROR_NOTIFICATION_TEMPLATE = "privacy_request_error_notification.html"
PRIVACY_REQUEST_REVIEW_DENY_TEMPLATE = "privacy_request_review_deny.html"
PRIVACY_REQUEST_REVIEW_APPROVE_TEMPLATE = "privacy_request_review_approve.html"
