.button {
    font-weight: 600;
    padding: 0 10px;
    border-radius: 6px;
    height: 32px;
    border: 0;
    cursor: pointer;
    font-size: 14px;
    font-family: 'Inter';

    &:disabled {
        opacity: 0.7;
        cursor: not-allowed;
    }
}

.primary {
    background-color: #5B301D;
    color: #fff;

    &:hover {
        background-color: #D78F3F;
    }
}

.secondary {
    background-color: #fff;
    border: 1px solid #E2E8F0;

    &:hover {
        background-color: #f5f5f5;
    }
}