"""privacy notice attributes

Revision ID: adad3be6af08
Revises: 7a94527643a1
Create Date: 2022-03-22 19:22:15.248311

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "adad3be6af08"
down_revision = "fdd28c39a679"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "data_subjects",
        sa.Column("rights", sa.JSON(), nullable=True),
    )
    op.add_column(
        "data_subjects",
        sa.Column("automated_decisions_or_profiling", sa.BOOLEAN(), nullable=True),
    )
    op.add_column(
        "data_uses", sa.Column("legitimate_interest", sa.BOOLEAN(), nullable=True)
    )
    op.add_column(
        "data_uses",
        sa.Column("legitimate_interest_impact_assessment", sa.String(), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("data_uses", "legitimate_interest_impact_assessment")
    op.drop_column("data_uses", "legitimate_interest")
    op.drop_column("data_subjects", "automated_decisions_or_profiling")
    op.drop_column("data_subjects", "rights")
    # ### end Alembic commands ###
