"""Add fidesctl_meta field to dataset

Revision ID: be432bd23596
Revises: adad3be6af08
Create Date: 2022-04-01 00:15:07.569104

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "be432bd23596"
down_revision = "a2b9b156ebaf"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("datasets", sa.Column("fidesctl_meta", sa.JSON(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("datasets", "fidesctl_meta")
    # ### end Alembic commands ###
