(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[4166],{42376:function(i,e,n){"use strict";n.r(e);var t=n(23256),s=n(44359),r=n(20440),a=n(41664),d=n.n(a),o=n(11163),u=n(67971),c=n(84247),l=n(57418),h=n(18459),v=n(85893);e.default=function(){var i=function(){var i=(0,o.useRouter)().query.id,e=void 0===i?"":i,n=(0,u.QA)({id:Array.isArray(e)?e[0]:e,verbose:!0},{skip:""===e});return{data:n.data,isLoading:n.isLoading,isUninitialized:n.isUninitialized}}(),e=i.data,n=i.isLoading,a=i.isUninitialized,f=e&&0!==(null===e||void 0===e?void 0:e.items.length)?(0,v.jsx)(l.Z,{data:null===e||void 0===e?void 0:e.items[0]}):(0,v.jsx)(t.xv,{children:"404 no privacy request found"});return(n||a)&&(f=(0,v.jsx)(t.M5,{children:(0,v.jsx)(s.$,{})})),(0,v.jsxs)(c.Z,{title:"Privacy Request",children:[(0,v.jsxs)(t.X6,{fontSize:"2xl",fontWeight:"semibold",children:["Privacy Request",(0,v.jsx)(t.xu,{mt:2,mb:9,children:(0,v.jsxs)(r.aG,{fontWeight:"medium",fontSize:"sm",children:[(0,v.jsx)(r.gN,{children:(0,v.jsx)(r.At,{as:d(),href:h.gp,children:"Privacy Requests"})}),(0,v.jsx)(r.gN,{children:(0,v.jsx)(r.At,{isCurrentPage:!0,color:"complimentary.500",_hover:{textDecoration:"none"},children:"View Details"})})]})})]}),f]})}},87519:function(i,e,n){(window.__NEXT_P=window.__NEXT_P||[]).push(["/subject-request/[id]",function(){return n(42376)}])}},function(i){i.O(0,[5372,7601,2890,2350,440,8522,5657,6508,7418,9774,2888,179],(function(){return e=87519,i(i.s=e);var e}));var e=i.O();_N_E=e}]);