(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[6355],{7907:function(r,e,n){"use strict";n.r(e),n.d(e,{default:function(){return O}});n(67294);var t=n(59499),c=n(16835),o=n(23256),i=n(10061),u=n(43072),s=n(53547),f=n(85893);function a(r,e){var n=Object.keys(r);if(Object.getOwnPropertySymbols){var t=Object.getOwnPropertySymbols(r);e&&(t=t.filter((function(e){return Object.getOwnPropertyDescriptor(r,e).enumerable}))),n.push.apply(n,t)}return n}function p(r){for(var e=1;e<arguments.length;e++){var n=null!=arguments[e]?arguments[e]:{};e%2?a(Object(n),!0).forEach((function(e){(0,t.Z)(r,e,n[e])})):Object.getOwnPropertyDescriptors?Object.defineProperties(r,Object.getOwnPropertyDescriptors(n)):a(Object(n)).forEach((function(e){Object.defineProperty(r,e,Object.getOwnPropertyDescriptor(n,e))}))}return r}var b=function(){var r=(0,u.ny)(),e=(0,c.Z)(r,1)[0];return(0,f.jsx)("div",{children:(0,f.jsxs)("main",{children:[(0,f.jsx)(o.X6,{mb:4,fontSize:"xl",colorScheme:"primary",children:"Profile"}),(0,f.jsx)(o.iz,{mb:7}),(0,f.jsx)(s.Z,{onSubmit:function(r){var n=(0,i.JJ)(r.password);return e(p(p({},r),{},{password:n}))}})]})})},l=n(33278),O=function(){return(0,f.jsx)(l.Z,{title:"New User",children:(0,f.jsx)(b,{})})}},51050:function(r,e,n){(window.__NEXT_P=window.__NEXT_P||[]).push(["/user-management/new",function(){return n(7907)}])}},function(r){r.O(0,[5372,7601,2890,2350,6211,440,6648,6508,6016,3382,9774,2888,179],(function(){return e=51050,r(r.s=e);var e}));var e=r.O();_N_E=e}]);