self.__BUILD_MANIFEST=function(s,a,e,c,t,n,i,d,u,f,r,b,o,h,j,k,g,m,p,y,l,_,w,q){return{__rewrites:{beforeFiles:[],afterFiles:[{source:"/api/v1/:path"},{source:"/api/v1/:first/:second*"},{source:"/health"}],fallback:[]},"/":[s,a,c,t,n,i,d,f,b,e,r,p,"static/chunks/pages/index-b164d4a8cf444e77.js"],"/404":["static/chunks/pages/404-4f91cbf004099ef5.js"],"/_error":["static/chunks/pages/_error-cd8de5d864b8d699.js"],"/add-systems":[s,a,c,t,n,i,d,f,b,o,k,e,r,y,"static/chunks/pages/add-systems-5cf54a9565c62d82.js"],"/classify-systems":[s,a,i,o,e,"static/chunks/pages/classify-systems-94ae3ec9aa881b2b.js"],"/dataset":[s,a,u,i,"static/chunks/pages/dataset-df2f929f2a077f06.js"],"/dataset/new":[s,a,c,t,n,u,f,g,e,r,"static/chunks/pages/dataset/new-5934dada28c2dddd.js"],"/dataset/[id]":[s,a,t,n,u,i,d,f,m,k,e,r,"static/chunks/pages/dataset/[id]-d23b79f7e8f3dbc3.js"],"/datastore-connection":[s,a,b,h,"static/css/684ec0de717c9d2d.css","static/chunks/pages/datastore-connection-3cda568ca32a75dd.js"],"/datastore-connection/new":[s,a,c,n,u,d,b,j,h,l,g,m,e,_,"static/chunks/pages/datastore-connection/new-f4803fa71900d06f.js"],"/datastore-connection/[id]":[s,a,c,n,u,d,b,j,h,l,g,m,e,_,"static/chunks/pages/datastore-connection/[id]-9f1e3444a7d28774.js"],"/login":["static/chunks/pages/login-196778433a92986b.js"],"/management/about":[s,t,"static/chunks/pages/management/about-c848cc197856a742.js"],"/privacy-requests":[s,a,c,t,n,i,d,f,e,r,p,"static/chunks/pages/privacy-requests-5c6c1593927eb918.js"],"/privacy-requests/[id]":[s,a,c,t,u,i,d,e,w,"static/chunks/pages/privacy-requests/[id]-f8e4ee7330532ee2.js"],"/subject-request/[id]":[s,a,c,t,u,i,d,e,w,"static/chunks/pages/subject-request/[id]-fbe3f7f2bc9581b4.js"],"/system":[s,h,e,"static/css/670d677a9586d067.css","static/chunks/pages/system-4d979f0e6407b7f7.js"],"/system/configure":[s,a,c,t,n,u,i,d,f,o,j,k,e,r,y,"static/chunks/pages/system/configure-0048f341d11f9b88.js"],"/taxonomy":[s,a,c,t,n,d,f,o,j,e,r,"static/chunks/pages/taxonomy-eb6b4b801ecff19d.js"],"/user-management":[s,i,"static/chunks/pages/user-management-35fef82d27392e29.js"],"/user-management/new":[s,a,c,t,n,u,f,e,r,q,"static/chunks/pages/user-management/new-959fa8d1fdfe2370.js"],"/user-management/profile/[id]":[s,a,c,t,n,u,f,e,r,q,"static/chunks/pages/user-management/profile/[id]-1138c7bdfb062d53.js"],sortedPages:["/","/404","/_app","/_error","/add-systems","/classify-systems","/dataset","/dataset/new","/dataset/[id]","/datastore-connection","/datastore-connection/new","/datastore-connection/[id]","/login","/management/about","/privacy-requests","/privacy-requests/[id]","/subject-request/[id]","/system","/system/configure","/taxonomy","/user-management","/user-management/new","/user-management/profile/[id]"]}}("static/chunks/5372-3ffae53d42cee606.js","static/chunks/7601-984deb12794e2fab.js","static/chunks/6508-2ee73dcfad69f357.js","static/chunks/2890-35aec91adba37a4d.js","static/chunks/2350-9ddbd589c864c3c8.js","static/chunks/6211-fbf5ef4bd9c6c367.js","static/chunks/8522-7ee39d346fb4144d.js","static/chunks/5657-b6ec86f07c3363de.js","static/chunks/440-3d124aad9abef645.js","static/chunks/6648-a52fe1c2b76114f7.js","static/chunks/6016-fc6c3c6c3e696ad9.js","static/chunks/4-66d823197570d84b.js","static/chunks/4530-204fe482e468bd6e.js","static/chunks/9816-dbbf4df71177c6c6.js","static/chunks/1144-9070ccde413c0dde.js","static/chunks/6049-45f81f8df7f5967b.js","static/chunks/1272-772e698a9da13171.js","static/chunks/2698-027e58560079638e.js","static/chunks/2980-82ac3e94eb00873e.js","static/chunks/5056-4eaf1ba4bb8c0908.js","static/chunks/6894-d3cc414886ca0568.js","static/chunks/4777-663dc0361379eb9f.js","static/chunks/7418-917877f2f553c91b.js","static/chunks/3382-d773f14cd040a0cd.js"),self.__BUILD_MANIFEST_CB&&self.__BUILD_MANIFEST_CB();