"""add_due_date

Revision ID: d8df7ff7aab4
Revises: 912d801f06c0
Create Date: 2022-09-06 15:21:21.181463

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d8df7ff7aab4"
down_revision = "912d801f06c0"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "privacyrequest",
        sa.Column("due_date", sa.DateTime(timezone=True), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("privacyrequest", "due_date")
    # ### end Alembic commands ###
