"""add column for special_category to data_uses

Revision ID: 7a94527643a1
Revises: 150f234ef1de
Create Date: 2022-03-18 03:22:23.975102

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "7a94527643a1"
down_revision = "150f234ef1de"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("data_uses", sa.Column("special_category", sa.Text(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("data_uses", "special_category")
    # ### end Alembic commands ###
